# coding: utf-8

import unittest
import yaml
import types
from snakemake.sourcecache import WorkflowError
from hydra_genetics.utils.misc import get_input_aligned_bam, get_input_haplotagged_bam


class TestResourcesUtils(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_merge(self):
        from hydra_genetics.utils.misc import merge
        dict1a = {'bwa_mem': {'extra': "some settings"}}
        dict1b = {'bwa_mem': {'extra': "some settings"}}
        dict2 = {'bwa_mem': {'cpu': 20, 'time': '12:00:00'}}
        dict1a.update(dict2)
        self.assertEqual({'bwa_mem': {'cpu': 20, 'time': '12:00:00'}}, dict1a)
        dict1b = merge(dict1b, dict2)
        self.assertEqual({'bwa_mem': {'extra': "some settings", 'cpu': 20, 'time': '12:00:00'}}, dict1b)

    def test_extract_chr(self):
        from hydra_genetics.utils.misc import extract_chr
        self.assertEqual(extract_chr("tests/utils/files/ref.fasta.fai"), [
                                     'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9', 'chr10', 'chr11',
                                     'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18', 'chr19', 'chr20', 'chr21',
                                     'chr22', 'chrX', 'chrY'])
        self.assertEqual(extract_chr("tests/utils/files/ref.fasta.fai", ['chr6', 'chr17', 'chrX']), [
                                     'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr7', 'chr8', 'chr9', 'chr10', 'chr11', 'chr12',
                                     'chr13', 'chr14', 'chr15', 'chr16', 'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrY',
                                     'chrM'])
        self.assertEqual(extract_chr(""), [''])

    def test_variable_replacement(self):
        with open("tests/utils/files/config_variable_replacement.yaml") as f:
            config = yaml.load(f, Loader=yaml.loader.SafeLoader)
        print(config)
        self.assertEqual(config,
                         {'PROJECT': 'MY variable', 'bwa_mem': {'extra': 'some settings {{PROJECT}}'}})

        from hydra_genetics.utils.misc import replace_dict_variables
        self.assertEqual(replace_dict_variables(config),
                         {'PROJECT': 'MY variable', 'bwa_mem': {'extra': 'some settings MY variable'}})


class TestGetInputAlignedBam(unittest.TestCase):
    def test_with_aligner(self):
        config = {"aligner": "minimap2"}
        wildcards = types.SimpleNamespace(sample="S1", type="T")
        bam, bai = get_input_aligned_bam(wildcards, config)
        self.assertEqual(bam, "alignment/minimap2_align/S1_T.bam")
        self.assertEqual(bai, "alignment/minimap2_align/S1_T.bam.bai")

    def test_without_aligner(self):
        config = {}
        wildcards = types.SimpleNamespace(sample="S2", type="N")
        bam, bai = get_input_aligned_bam(wildcards, config)
        self.assertEqual(bam, "alignment/samtools_merge_bam/S2_N.bam")
        self.assertEqual(bai, "alignment/samtools_merge_bam/S2_N.bam.bai")

    def test_custom_default_path(self):
        config = {}
        wildcards = types.SimpleNamespace(sample="S3", type="T")
        bam, bai = get_input_aligned_bam(wildcards, config, default_path="custom/path")
        self.assertEqual(bam, "custom/path/S3_T.bam")
        self.assertEqual(bai, "custom/path/S3_T.bam.bai")

    def test_missing_sample_key(self):
        config = {}
        wildcards = types.SimpleNamespace(type="T")
        with self.assertRaises(Exception):
            get_input_aligned_bam(wildcards, config)

    def test_missing_type_key(self):
        config = {}
        wildcards = types.SimpleNamespace(sample="S4")
        with self.assertRaises(Exception):
            get_input_aligned_bam(wildcards, config)


class TestGetInputHaplotaggedBam(unittest.TestCase):
    def test_default_path_only(self):
        wildcards = types.SimpleNamespace(sample="S10", type="N")
        config = dict()
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config,
            default_path="custom/default/path"
        )
        self.assertEqual(bam, "custom/default/path/S10_N.bam")
        self.assertEqual(bai, "custom/default/path/S10_N.bam.bai")

    def test_missing_type_wildcard(self):
        wildcards = types.SimpleNamespace(sample="S11")
        config = dict()
        with self.assertRaises(WorkflowError):
            get_input_haplotagged_bam(wildcards, config)

    def test_missing_sample_wildcard(self):
        wildcards = types.SimpleNamespace(type="T")
        config = dict()
        with self.assertRaises(WorkflowError):
            get_input_haplotagged_bam(wildcards, config)

    def test_empty_haplotag_path(self):
        wildcards = types.SimpleNamespace(sample="S12", type="T")
        config = dict({'haplotag_path': ""})
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config
        )
        self.assertEqual(bam, "S12_T.bam")
        self.assertEqual(bai, "S12_T.bam.bai")

    def test_missing_haplotag_path(self):
        wildcards = types.SimpleNamespace(sample="S14", type="N")
        config = dict()
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config
        )
        self.assertEqual(bam, "alignment/samtools_merge_bam/S14_N.bam")
        self.assertEqual(bai, "alignment/samtools_merge_bam/S14_N.bam.bai")

    def test_default_values(self):
        wildcards = types.SimpleNamespace(sample="S13", type="N")
        config = dict()
        bam, bai = get_input_haplotagged_bam(wildcards, config)
        self.assertEqual(bam, "alignment/samtools_merge_bam/S13_N.bam")
        self.assertEqual(bai, "alignment/samtools_merge_bam/S13_N.bam.bai")

    def test_with_custom_suffix(self):
        wildcards = types.SimpleNamespace(sample="S5", type="T")
        config = dict({"haplotag_path": "snv_indels/whatshap_haplotag"})
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config,
            suffix="custom_suffix"
        )
        self.assertEqual(bam, "snv_indels/whatshap_haplotag/S5_T.custom_suffix.bam")
        self.assertEqual(bai, "snv_indels/whatshap_haplotag/S5_T.custom_suffix.bam.bai")

    def test_with_suffix_in_config(self):
        wildcards = types.SimpleNamespace(sample="S7", type="N")
        config = dict({
            "haplotag_path": "snv_indels/whatshap_haplotag",
            "haplotag_suffix": "from_config"
        })
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config
        )
        self.assertEqual(bam, "snv_indels/whatshap_haplotag/S7_N.from_config.bam")
        self.assertEqual(bai, "snv_indels/whatshap_haplotag/S7_N.from_config.bam.bai")

    def test_with_empty_suffix(self):
        wildcards = types.SimpleNamespace(sample="S6", type="T")
        config = dict({"haplotag_path": "snv_indels/whatshap_haplotag"})
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config,
            suffix=""
        )
        self.assertEqual(bam, "snv_indels/whatshap_haplotag/S6_T.bam")
        self.assertEqual(bai, "snv_indels/whatshap_haplotag/S6_T.bam.bai")

    def test_with_all_parameters(self):
        wildcards = types.SimpleNamespace(sample="S8", type="T")
        config = dict({"haplotag_path": "custom/path"})
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config,
            default_path="should/not/use",
            suffix="special"
        )
        self.assertEqual(bam, "custom/path/S8_T.special.bam")
        self.assertEqual(bai, "custom/path/S8_T.special.bam.bai")

    def test_none_suffix(self):
        wildcards = types.SimpleNamespace(sample="S9", type="N")
        config = dict({"haplotag_path": "custom/path"})
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config,
            suffix=None
        )
        self.assertEqual(bam, "custom/path/S9_N.bam")
        self.assertEqual(bai, "custom/path/S9_N.bam.bai")

    def test_with_aligner_no_haplotag_path(self):
        wildcards = types.SimpleNamespace(sample="S1", type="T")
        config = {"aligner": "bwa-mem2"}
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config
        )
        self.assertEqual(bam, "alignment/bwa-mem2_align/S1_T.bam")
        self.assertEqual(bai, "alignment/bwa-mem2_align/S1_T.bam.bai")

    def test_haplotag_path_takes_precedence_over_aligner(self):
        wildcards = types.SimpleNamespace(sample="S1", type="T")
        config = {
            "aligner": "bwa-mem2",
            "haplotag_path": "snv_indels/whatshap_haplotag"
        }
        bam, bai = get_input_haplotagged_bam(
            wildcards,
            config
        )
        self.assertEqual(bam, "snv_indels/whatshap_haplotag/S1_T.bam")
        self.assertEqual(bai, "snv_indels/whatshap_haplotag/S1_T.bam.bai")



import shutil
import tempfile
import unittest
from requests import HTTPError

from hydra_genetics.utils.io.reference import fetch_reference_data, fetch_url_content


class TestReference(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.outdir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        super().tearDown()
        shutil.rmtree(self.tmpdir)
        shutil.rmtree(self.outdir)

    def test_fetch_url_content_404(self):
        url = "https://github.com/Twist_Solid_pipeline_files/raw/v1.0.0/design/this-is-surely-not-a-file.txt"
        with self.assertRaises(HTTPError) as e:
            fetch_url_content(url, self.outdir, self.tmpdir)
        self.assertEqual(e.exception.status, 404)

    def test_fetch_reference_data_404(self):
        validation_data = {
            "not a file": {
                "path": "files/not-a-file.txt",
                "type": "file",
                "url": "https://github.com/Twist_Solid_pipeline_files/raw/v1.0.0/design/this-is-surely-not-a-file.txt",
            },
        }
        fetched, links, failed, skipped = fetch_reference_data(validation_data, self.outdir)
        self.assertEqual(len(fetched), 0)
        self.assertEqual(len(failed), 1)
        self.assertEqual(len(links), 0)
        self.assertEqual(len(skipped), 0)



import re

_chr_pattern = re.compile(r'^chr[XYM0-9]+$|^[XYM0-9]+$')
_nc_pattern = re.compile(r'^NC_0+\d+\.\d+$')


class ChrTranslater(object):
    def __init__(self, mapper_file):
        self.chr_to_nc = dict()
        self.nc_to_chr = dict()

        with open(mapper_file) as mapping:
            for line in mapping:
                if not line.startswith("#"):
                    columns = line.split("\t")
                    if _chr_pattern.match(columns[0]) and _nc_pattern.match(columns[1]):
                        self.chr_to_nc[columns[0]] = columns[1]
                        self.nc_to_chr[columns[1]] = columns[0]
                    elif _chr_pattern.match(columns[1]) and _nc_pattern.match(columns[0]):
                        self.chr_to_nc[columns[1]] = columns[0]
                        self.nc_to_chr[columns[0]] = columns[1]
                    else:
                        raise Exception("Unexpected column values for column 1/2: {}".format(line))

    def get_chr_value(self, nc_id):
        return self.nc_to_chr[nc_id]

    def get_nc_value(self, chr_id):
        return self.chr_to_nc[chr_id]



# coding: utf-8

from collections.abc import Mapping
from copy import deepcopy
import os
import yaml
import re
from datetime import datetime
from snakemake.sourcecache import GithubFile, LocalGitFile, WorkflowError


def get_module_snakefile(config, repo, path, tag):
    ''' return a complete snakefile path, either pointing to github or a local repo'''
    if 'hydra_local_path' in config:
        repo_path = os.path.join(config['hydra_local_path'], repo)
        if not os.path.exists(repo_path):
            ' Remove possible organization part from repo path '
            repo_path_v2 = os.path.join(config['hydra_local_path'], os.sep.join(repo.split(os.sep)[1:]))
            if not os.path.exists(repo_path_v2):
                raise FileNotFoundError(f"Repo folder {repo_path} or {repo_path_v2} doesn't exist!")
            else:
                repo_path = repo_path_v2
        if not os.path.isdir(repo_path):
            raise FileNotFoundError(f"{repo_path} isn't a folder!")
        return LocalGitFile(repo_path, path=path, tag=tag)
    else:
        return GithubFile(repo, path=path, tag=tag)


def merge(dict1, dict2):
    ''' Return a new dictionary by merging two dictionaries recursively. '''

    result = deepcopy(dict1)
    for key, value in dict2.items():
        if isinstance(value, Mapping):
            result[key] = merge(result.get(key, {}), value)
        else:
            result[key] = deepcopy(dict2[key])
    return result


def extract_chr(file, filter_out=["chrM"]):
    chr = None
    if os.path.exists(file):
        with open(file) as lines:
            chr = [line.split("\t")[0] for line in lines]
        return [c for c in chr if c not in filter_out]
    return [""]


def replace_dict_variables(config):
    '''
        Make it possible to defina a variable in a  vaue string using ´{{}}´,
        ex {{PROJECT_VARIABLE}}. PROJECT_VARIABLE must exist as a variable
        in the yaml file and will be used to replace all occurences of
        {{PROJECT_VARAIBLE}} in the yaml file.
    '''
    def update_dict(temp_config):
        for k in temp_config:
            if type(temp_config[k]) is dict:
                temp_config[k] = update_dict(temp_config[k])
            elif type(temp_config[k]) is str:
                match = re.findall(r'\{\{([A-Za-z0-9_]+)\}\}', temp_config[k])
                for m in match:
                    temp_config[k] = temp_config[k].replace("{{" + m + "}}", config[m])
            elif type(temp_config[k]) is list:
                for i in range(len(temp_config[k])):
                    if type(temp_config[k][i]) is str:
                        match = re.findall(r'\{\{([A-Za-z0-9_]+)\}\}', temp_config[k][i])
                        for m in match:
                            temp_config[k][i] = temp_config[k][i].replace("{{" + m + "}}", config[m])
        return temp_config
    return update_dict(config)


def export_config_as_file(config, output_file="config", directory="versions", date_string=None):
    if date_string is None:
        date_string = datetime.now().strftime('%Y%m%d--%H-%M-%S')
    if directory is not None and len(directory) > 0:
        date_string = f"_{date_string}"
    if output_file is not None:
        output_file = f"{output_file}{date_string}.yaml"
    if directory is not None:
        output_file = os.path.join(directory, output_file)
    if len(directory) > 0 and not os.path.isdir(directory):
        os.makedirs(directory)
    with open(output_file, 'w') as writer:
        writer.write(yaml.dump(config))


def get_input_aligned_bam(wildcards, config, default_path="alignment/samtools_merge_bam"):
    """
    Compile the paths to input aligned BAM and BAI files for the workflow.

    This function determines the appropriate BAM file path based on the configuration
    and workflow parameters.

    Args:
        wildcards (snakemake.io.Wildcards): Wildcards object containing sample and type information.
        config (dict): Configuration dictionary with possible keys:
            - aligner (str or None): The aligner used for generating the BAM file.
        default_path (str): Default path for BAM files if no specific configuration is provided.

    Returns:
        tuple: A tuple containing the alignment BAM file path and its BAI index file path.
    """
    try:
        if config.get("aligner") is not None:
            # Use aligner to compile the paths
            aligner = config.get("aligner")
            alignment_path = f"alignment/{aligner}_align/{wildcards.sample}_{wildcards.type}.bam"
            index_path = f"alignment/{aligner}_align/{wildcards.sample}_{wildcards.type}.bam.bai"
        else:
            # no aligner, use the default path
            alignment_path = f"{default_path}/{wildcards.sample}_{wildcards.type}.bam"
            index_path = f"{alignment_path}.bai"

        return alignment_path, index_path

    except KeyError as e:
        raise WorkflowError(f"Missing required wildcards: {e}")


def get_input_haplotagged_bam(wildcards, config, default_path="alignment/samtools_merge_bam", suffix=None):
    """
    Compile paths to haplotagged BAM/BAI files (may be required for downstream analyses with e.g. cnvkit_batch)
    This function determines the appropriate BAM file path based on the configuration. suffix can be provided
    either as a function argument or in the config file under the key 'haplotag_suffix'.

    For backward compatibility, if only wildcards and config are provided, the function will default to using the
    default path and return 'alignment/samtools_merge_bam/{sample}_{type}.bam'.

    Args:
        wildcards (snakemake.io.Wildcards): Wildcards object containing sample and type information.
        config (dict): config with or without 'haplotag_path' key.
        default_path (str): Default path for BAM files if no specific configuration is provided.
        suffix (str): Suffix to append to the BAM file name (default is None).
    Returns:
        tuple: A tuple containing the alignment BAM file path and its BAI index file path.

    """
    try:
        sample_name = getattr(wildcards, "sample")
        sample_type = getattr(wildcards, "type")
    except AttributeError as e:
        raise WorkflowError(f"Missing required wildcards: {e}")

    haplotag_path = config.get("haplotag_path", None)
    if haplotag_path is not None:
        path_to_input_bam = haplotag_path
    elif config.get("aligner") is not None:
        aligner = config.get("aligner")
        path_to_input_bam = f"alignment/{aligner}_align"
    else:
        path_to_input_bam = default_path

    # check for suffix in the config
    if suffix is None:
        suffix = config.get("haplotag_suffix", None)

    if suffix:  # This will be False for None and ''
        file_name = f"{sample_name}_{sample_type}.{suffix}.bam"
    else:
        file_name = f"{sample_name}_{sample_type}.bam"

    bam_path = os.path.join(path_to_input_bam, file_name)
    bai_path = f"{bam_path}.bai"

    return bam_path, bai_path



import click
import logging
import os
import yaml


def create_docker_file_path(storage_path, container):
    return os.path.join(storage_path, container.replace('docker://', '').replace("/", "__").replace(":", "_") + ".sif")


@click.group("prepare-environment", short_help="prepare for singularity environment")
def environment():
    pass


@environment.command(short_help="update config with path to container cache")
@click.option(
    "-c",
    "--configfile",
    prompt="config.yaml path",
    required=True,
    type=str,
    help="config file used to find pipeline containers",
)
@click.option(
    "-n",
    "--new-configfile",
    prompt="path",
    required=True,
    type=str,
    help="path to update config file",
)
@click.option(
    "-p",
    "--singularity-cache-path",
    prompt="singularity cache path",
    required=True,
    type=str,
    help="path to where singulariy files are stored",
)
def container_path_update(configfile, new_configfile, singularity_cache_path):
    with open(configfile, "r") as stream:
        try:
            yaml_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    for key in yaml_data:
        if 'container' in key:
            yaml_data[key] = create_docker_file_path(singularity_cache_path, yaml_data[key])
        elif isinstance(yaml_data[key], dict):
            if 'container' in yaml_data[key]:
                yaml_data[key]['container'] = create_docker_file_path(singularity_cache_path, yaml_data[key]['container'])

    with open(new_configfile, 'w') as file:
        for key, value in yaml_data.items():
            file.write(yaml.dump({key: value}))
            file.write("\n")


@environment.command(short_help="update search path to reference files")
@click.option(
    "-c",
    "--configfile",
    prompt="config.yaml path",
    required=True,
    type=str,
    help="config file used to find pipeline containers",
)
@click.option(
    "-n",
    "--new-configfile",
    prompt="path",
    required=True,
    type=str,
    help="path to update config file",
)
@click.option(
    "--reference-path",
    multiple=True,
    prompt="path",
    required=True,
    type=str,
    help=(
        "format oldpath:newpath, will iterate over the provided config"
        " file and replace all occurrences of oldpath with new path"
    ),
)
def reference_path_update(configfile, new_configfile, reference_path):
    def process_yaml_data(yaml_data, reference_path):
        for key in yaml_data:
            if isinstance(yaml_data[key], dict):
                yaml_data[key] = process_yaml_data(yaml_data[key], reference_path)
            else:
                if isinstance(yaml_data[key], str):
                    for old_path, new_path in reference_path:
                        if old_path in yaml_data[key]:
                            yaml_data[key] = yaml_data[key].replace(old_path, new_path)
        return yaml_data
    reference_paths = [r.split(":") for r in reference_path]
    with open(configfile, "r") as stream:
        try:
            yaml_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    yaml_data = process_yaml_data(yaml_data, reference_paths)

    with open(new_configfile, 'w') as file:
        for key, value in yaml_data.items():
            file.write(yaml.dump({key: value}))
            file.write("\n")


@environment.command(short_help="create singularity containers")
@click.option(
    "-c",
    "--configfile",
    prompt="config.yaml path",
    required=True,
    type=str,
    help="config file used to find pipeline containers",
)
@click.option(
    "-o",
    "--output-folder",
    prompt="folder path",
    required=True,
    type=str,
    help="output folder where files will be stored",
)
def create_singularity_files(configfile, output_folder):
    yaml_data = None
    with open(configfile, "r") as stream:
        try:
            yaml_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            exit(1)

    containers = set()
    for key in yaml_data:
        if "container" in key:
            containers.add(yaml_data[key])
        elif isinstance(yaml_data[key], dict):
            if 'container' in yaml_data[key]:
                containers.add(yaml_data[key]['container'])

    if os.path.exists(output_folder):
        if os.path.isdir(output_folder):
            pass
        else:
            raise IOError(f"{output_folder} is not a folder")
    else:
        os.mkdir(output_folder)

    for container in containers:
        file_path = create_docker_file_path(output_folder, container)
        logging.info(f"Create singularity {file_path}")
        os.system(f"singularity build {file_path} {container}")



# vim: syntax=python tabstop=4 expandtab
# coding: utf-8

import logging
import os
import shutil
import tempfile
import unittest

logger = logging.getLogger(__name__).addHandler(logging.NullHandler())


class TestModels(unittest.TestCase):
    def setUp(self):
        # create fixtures
        self.tempdir = tempfile.mkdtemp()
        with open(os.path.join(self.tempdir, "mapping"), 'w') as hotspots:
            hotspots.write("NC_000001.10\tchr1\t115252349\n")
            hotspots.write("NC_000017.10\tchr17\t37881164\n")
            hotspots.write("NC_000012.11\tchr12\t25398279")

    def tearDown(self):
        # delete fixtures
        shutil.rmtree(self.tempdir)

    def test_hotspot_creation(self):
        from hydra_genetics.utils.models.hotspot import Hotspot
        from hydra_genetics.utils.models.hotspot import ReportClass
        hotspot = Hotspot("NC_000017.10",
                          37880990,
                          37880990,
                          "ERBB2",
                          "c.123",
                          "p.V773",
                          ReportClass["region_all"], "-", "exon20", "NM_004448")
        self.assertEqual(hotspot.CHROMOSOME, "NC_000017.10")
        self.assertEqual(hotspot.START, 37880990)
        self.assertEqual(hotspot.END, 37880990)
        self.assertEqual(hotspot.GENE, "ERBB2")
        self.assertEqual(hotspot.CDS_MUTATION_SYNTAX, "c.123")
        self.assertEqual(hotspot.AA_MUTATION_SYNTAX, "p.V773")
        self.assertEqual(hotspot.REPORT, ReportClass.region_all)
        self.assertEqual(hotspot.COMMENT, "-")
        self.assertEqual(hotspot.EXON, "exon20")
        self.assertEqual(hotspot.ACCESSION_NUMBER, "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    "37880990",
                    37880990,
                    "ERBB2",
                    "c.123",
                    "p.V773",
                    ReportClass["region_all"], "-", "exon20", "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    37880990,
                    "37880990",
                    "ERBB2",
                    "c.123",
                    "p.V773",
                    ReportClass["region_all"], "-", "exon20", "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    37880990,
                    37880989,
                    "ERBB2",
                    "c.123",
                    "p.V773",
                    ReportClass["region_all"], "-", "exon20", "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    37880990,
                    37880990,
                    "ERBB2",
                    "p.123",
                    "p.V773",
                    ReportClass["region_all"], "-", "exon20", "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    37880990,
                    37880990,
                    "ERBB2",
                    "c.123",
                    "c.773",
                    ReportClass["region_all"], "-", "exon20", "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    37880990,
                    37880990,
                    "ERBB2",
                    "c.123",
                    "p.V773",
                    "region_all",
                    "-",
                    "exon20",
                    "NM_004448")

        with self.assertRaises(ValueError):
            Hotspot("NC_000017.10",
                    37880990,
                    37880990,
                    "ERBB2",
                    "c.123",
                    "p.V773",
                    ReportClass["region_all"], "-", "e20", "NM_004448")

    def test_mapping_creation(self):
        from hydra_genetics.utils.io.chr import ChrTranslater
        translater = ChrTranslater(os.path.join(self.tempdir, "mapping"))

        self.assertEqual("chr1", translater.get_chr_value("NC_000001.10"))
        self.assertEqual("chr12", translater.get_chr_value("NC_000012.11"))
        self.assertEqual("chr17", translater.get_chr_value("NC_000017.10"))

        self.assertEqual("NC_000001.10", translater.get_nc_value("chr1"))
        self.assertEqual("NC_000012.11", translater.get_nc_value("chr12"))
        self.assertEqual("NC_000017.10", translater.get_nc_value("chr17"))


if __name__ == '__main__':
    import logging
    import sys
    logging.basicConfig(level=logging.CRITICAL, stream=sys.stdout, format='%(message)s')
    unittest.main()



# vim: syntax=python tabstop=4 expandtab
# coding: utf-8

__author__ = "{{ author }}"
__copyright__ = "Copyright {{ year }}, {{ author }}"
__email__ = "{{ email }}"
__license__ = "GPL-3"


def test_dummy():
    from dummy import dummy
    assert dummy() == 1



import gzip
import hashlib
import logging
import os
import shutil
import tarfile
import tempfile
import requests
import time
from requests import HTTPError
from urllib.parse import urlparse

# Expected input file format
# --------------------------
# rule_name1:
#     variable1:
#        entry
#
# rule_name2:
#     variable2:
#        entry
#
#
# --------------------------
#
# Expected formats for an entry
# A file
# {checksum: md5sum_value, type: file, path: path_to_file, url: url_path}
# Compressed file that will be extracted
# {checksum: md5sum_value, compressed_checksum: md5sum_value, type: file, path: path_storage_location, url: url_path}
# Folder (it must be compressed)
# {type: folder, compressed_checksum: md5sum_value, path: folder_storage_location,
#   content_checksum: { file1:md5sum_value, file2:md5sum_value}, url: url_path}
# content_checksum is optional, when it's not set the only validation made is the existence of the folder


def fetch_reference_data(validation_data, output_dir,
                         fetched=[], links=[], failed=[], skipped=[],
                         force=False):
    '''
        Function used to validate entries found in validation files

        Parameters:
            validation_data (dict):
            output_dir (str): path to where data will be stored
            fetched (list): list with files that have been retrieved.
            links (list): list with links that have been retrieved.
            failed (list): list with files that couldn't correctly be retrieved
            skipped (list): list with files that weren't retrieved, checksum haven't changed

        Returns:
            (list, list, list, list): (fetched files, linked files, failed files, skipped files)
    '''
    for k, value in validation_data.items():
        # If validation entry
        if "path" in value:
            content_path = os.path.join(output_dir, value['path'])
            content_type = value['type']

            # Some entries may not be fetchable and only added for validation
            if 'url' not in value:
                logging.debug(f"Content {content_path} contains no url entry, i.e will not be retrieved!")
                continue

            # Check if content needs to be updated
            update_needed = update_needed_for_entry(value, output_dir)

            if update_needed or force:
                logging.info(f"updating {content_path}")
                with tempfile.TemporaryDirectory() as tmpdirname:
                    temp_content_holder = os.path.join(tmpdirname, "tempfile")

                    # Fetch content and merge any split files
                    try:
                        if not fetch_url_content(value['url'], temp_content_holder, tmpdirname):
                            failed.append(content_path)
                            continue
                    except HTTPError as e:
                        logging.error(f"failed to fetch resource at {value['url']}: {e}")
                        failed.append(content_path)
                        continue

                    checksum_value = value['compressed_checksum'] if 'compressed_checksum' in value else value['checksum']

                    if not checksum_validate_file(temp_content_holder, checksum_value, content_path):
                        failed.append(content_path)
                    else:
                        compressed_folder = None
                        if 'folder' in content_type:
                            if 'compressed_checksum' in value and 'content_checksum' in value:
                                parts = list(value['content_checksum'].keys())[0].split(os.sep)
                                if len(parts) > 1:
                                    compressed_folder = parts[0]

                        if 'compressed_checksum' in value:
                            if extract_compressed_data(value, content_type, content_path, temp_content_holder, compressed_folder, force):
                                logging.info(f"folder {content_path} retrieved")
                                fetched.append(value['path'])
                            else:
                                failed.append(content_path)
                        else:
                            logging.info(f"file {content_path} retrieved")
                            fetched.append(value['path'])
                            move_content(temp_content_holder, content_path)
            else:
                logging.debug(f"skipped {content_path}")
                skipped.append(content_path)

            if 'link' in value:
                link_path = os.path.join(output_dir, value['link'])

                if not os.path.lexists(link_path):
                    logging.info(f"Creating link: {link_path}")
                    content_path = os.path.abspath(os.path.join(output_dir, value['path']))
                    links.append(link_path)
                    link_path = os.path.abspath(link_path)
                    if not os.path.isdir(os.path.dirname(link_path)):
                        os.mkdir(os.path.dirname(link_path))
                    relpath = os.path.relpath(os.path.dirname(content_path), os.path.dirname(link_path))
                    if relpath == ".":
                        content_path = os.path.basename(content_path)
                    elif relpath == "..":
                        content_path = os.path.join(relpath, os.path.basename(content_path))
                    else:
                        content_path = os.path.join(output_dir, value['path'])
                    os.symlink(content_path, link_path)
                else:
                    logging.info(f"link already exist: {link_path}")
        else:
            # Nested entry, recursively process content
            fetched, links, failed, skipped = fetch_reference_data(validation_data[k],
                                                                   output_dir,
                                                                   fetched,
                                                                   links,
                                                                   failed,
                                                                   skipped,
                                                                   force)
    return fetched, links, failed, skipped


def fetch_url_content(url, content_holder, tmpdir):
    """
    Fetch content from the provided url and make sure that the downloaded file
    has the correct md5 value.
    """
    def download_with_retry(target_url, target_path):
        parsed_url = urlparse(target_url)
        base_url = f"{parsed_url.scheme}://{parsed_url.netloc}/"
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Referer": base_url,
            "Accept": "*/*"
        }
        with requests.Session() as s:
            s.headers.update(headers)
            for i in range(10):
                r = s.get(target_url, stream=True, allow_redirects=True, timeout=60)
                if r.status_code == 202:
                    logging.info(f"Figshare is preparing file (202). Waiting 10s... (Attempt {i+1})")
                    r.close()
                    time.sleep(10)
                    continue
                try:
                    r.raise_for_status()
                except HTTPError as e:
                    e.status = e.response.status_code
                    raise e
                with open(target_path, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024*1024):
                        if chunk:
                            f.write(chunk)
                return True
        return False

    if isinstance(url, dict):
        counter = 1
        list_of_temp_files = []
        for part_url, part_checksum in url.items():
            temp_file = os.path.join(tmpdir, f"file{counter}")
            list_of_temp_files.append(temp_file)
            if not download_with_retry(part_url, temp_file):
                return False
            if not checksum_validate_file(temp_file, part_checksum):
                logging.info(f"Failed to retrieved part {counter}: {part_url}, expected {part_checksum}")
                return False
            counter += 1
        with open(content_holder, 'wb') as writer:
            for temp_content in list_of_temp_files:
                with open(temp_content, 'rb') as reader:
                    shutil.copyfileobj(reader, writer)
        return True
    else:
        return download_with_retry(url, content_holder)


def validate_reference_data(validation_data, path_to_ref_data,
                            file_list=[], not_found_in_config=[], found=[],
                            pass_list=[], link_list=[], failed_list=[]):
    '''
        Validate all entries in the validation dict

        Parameters:
            validation_data (dict): load json/yaml file in dict format
                                    ex {key: {key_s1: {path: value}, key_s2: {path: value}, k2: {path: value}}
            path_to_ref_data (str): path to where data is stored
            file_list (list): list with files in config file
            not_found_in_config (list): files in loaded validation that couldn't be found in the configuration file
            pass_list (list): files/folders passed validation
            linked_list (list): defined and existing links
            failed_list (list): files/folders failed validation

        Returns:
            (list, list, list, list, list,):
                list with files that haven't been validated,
                reference files not found in config,
                files found and defined in config,
                files passed validation,
                links defined and existing
                files failed validation
    '''
    def track_files(path, file_list=[], not_found_in_config=[], found=[]):
        if path in found:
            pass
        else:
            file_found = []
            for f in file_list:
                if f.endswith(path) or f == path:
                    file_found.append(f)
                    break
                elif path in f:
                    file_found.append(f)
            if file_found:
                for f in file_found:
                    file_list.remove(f)
                found.append(path)
            else:
                for f in found:
                    if f.endswith(f) or f in path:
                        return file_list, not_found_in_config, found
                if path not in found:
                    not_found_in_config.append(path)
        return file_list, not_found_in_config, found

    def valid_link(output_dir, item):
        if 'link' in item:
            link_path = os.path.join(path_to_ref_data, item['link'])
            if not os.path.lexists(link_path):
                logging.info(f"Missing link: {link_path}")
                return False
        return True

    for k, item in validation_data.items():
        # A validation entry
        if "path" in item:
            file_path = os.path.join(path_to_ref_data, item['path'])
            if not os.path.exists(file_path):
                logging.error(f"Could not locate: {file_path}")
                failed_list.append(item['path'])
                file_list, not_found_in_config, found = track_files(item['path'], file_list, not_found_in_config, found)
            else:
                if 'checksum' not in item and "content_checksum" not in item:
                    if not valid_link(path_to_ref_data, item):
                        failed_list.append(item['path'])
                    else:
                        if 'link' in item:
                            link_list.append(item['link'])
                        pass_list.append(item['path'])
                        logging.debug(f"{item['path']} found, no checksum validation!")
                elif "content_checksum" in item:
                    logging.debug(f"Folder found: {item['path']}")
                    _, failed, not_found = checksum_validate_content(item['content_checksum'], file_path)
                    if not failed and not not_found:
                        pass_list.append(item['path'])
                    else:
                        failed_list.append(item['path'])
                else:
                    if checksum_validate_file(file_path, validation_data[k]['checksum']) and valid_link(path_to_ref_data, item):
                        if 'link' in item:
                            link_list.append(item['link'])
                        pass_list.append(item['path'])
                    else:
                        failed_list.append(item['path'])
                file_list, not_found_in_config, found = track_files(item['path'], file_list, not_found_in_config, found)
        else:
            # Nested entry, recursively process content
            (file_list, not_found_in_config, found,
             pass_list, link_list, failed_list) = validate_reference_data(validation_data[k],
                                                                          path_to_ref_data,
                                                                          file_list, not_found_in_config,
                                                                          found, pass_list, link_list, failed_list)
    return file_list, not_found_in_config, found, pass_list, link_list, failed_list


def update_needed_for_entry(item, parent_dir="./"):
    """
        Checks if entry needs to be update, i.e checksum has change or the file/folder doesn't exist

        Patameters:
            item (dict): see top part of this page
            parent_dir (string): path to where content will be saved

        Returns:
            bool: if the entry needs to be updated
    """
    if item['path'].startswith("/"):
        content_path = item['path']
    else:
        content_path = os.path.join(parent_dir, item['path'])

    if 'file' in item['type'] and not os.path.isfile(content_path) or \
       'folder' in item['type'] and not os.path.isdir(content_path):
        logging.debug(f"{content_path} not found")
        return True
    else:
        if "checksum" in item:
            return not checksum_validate_file(content_path, item['checksum'])
        elif "content_checksum" in item:
            _, failed, not_found = checksum_validate_content(item['content_checksum'], content_path)
            if failed or not_found:
                return True
            else:
                return False
        else:
            logging.debug(f"{content_path} not validation done")


def checksum_validate_content(file_checksums, parent_dir=None, print_path_name=None, ) -> (int, int, int):
    """
        Used to validate content of a folders using a dict with path:md5sum values.

        Parameters:
            file_checksum (dict): dict with the following format {file_path: md5sum_value, }
            parent_dir (string): path to where the file is located
            print_path_name (string): when storing a file/folder using a temporary name this makes it possible to
                                      print a more human readable string.

        Returns:
            (int, int, int): file that passed checksum validation, files that failed checksum validation, files not found
    """
    passed = 0
    failed = 0
    not_found = 0
    failed_files = []
    files_not_found = []
    for file, checksum in file_checksums.items():
        full_file_path = os.path.join(parent_dir if parent_dir is not None else "./", file)
        if os.path.isfile(full_file_path):
            if checksum_validate_file(
                os.path.join(full_file_path),
                checksum,
                file
            ):
                passed += 1
            else:
                failed_files.append(file)
                failed += 1
        else:
            logging.debug(f"{full_file_path} not found")
            files_not_found.append(file)
            not_found += 1
    logging.debug(f"folder content validation, valid: {passed}, invalid {failed}, missing {not_found}")
    logging.debug(f"failed files: {failed_files}")
    logging.debug(f"failed files: {files_not_found}")
    return passed, failed, not_found


def checksum_validate_file(file, expected_checksum, print_path_name=None) -> bool:
    """
        function used to validate checksum.

        Parameters:
            file (str): path to file
            expected_checksum (str): md5sum value
            print_path_name (str): when storing a file/folder using a temporary name this makes it possible to
                                   print a more human readable string.

        Returns
            bool: true passed validation
    """
    m = hashlib.md5()
    with open(file, 'rb') as fp:
        for chunk in fp:
            m.update(chunk)
    calculated_md5 = m.hexdigest()
    print_name = print_path_name if print_path_name is not None else file
    if calculated_md5 == expected_checksum:
        logging.debug(f"{print_name}: valid checksum")
        return True
    else:
        logging.error(f"{print_name}: invalid checksum, got {calculated_md5}, expected {expected_checksum}")
        return False


def extract_compressed_data(item, content_type, content_path, temp_content_holder, compressed_folder=None, force=False) -> bool:
    """
        Function used to extract compressed content, both files and folders.

        Parameters:
            item (dict): a dict with checksum or content_checksum (a dict with path:md5sum values)
            content_path (string): path to where the file or folder will be saved
            temp_content_folder (string): path to where content can be temporary stored
            compress_folder (string): name of folder that should be moved.

        Returns:
            bool: if extraction was succefull
    """
    extracted_content_path = f"{temp_content_holder}_extracted"
    if "file" in content_type:
        with gzip.open(temp_content_holder, 'rb') as file:
            with open(extracted_content_path, 'wb') as writer:
                for line in file:
                    writer.write(line)
        if checksum_validate_file(extracted_content_path, item['checksum']):
            move_content(extracted_content_path, content_path)
            return True
        else:
            return False
    else:
        with tarfile.open(temp_content_holder, mode="r|gz") as tar:
            tar.extractall(extracted_content_path)
        if 'content_checksum' in item:
            if checksum_validate_content(item['content_checksum'], extracted_content_path, content_path):
                if compressed_folder is not None:
                    extracted_content_path = os.path.join(extracted_content_path, compressed_folder)
                    content_path = os.path.join(content_path, compressed_folder)
                move_content(extracted_content_path, content_path)
                return True
            else:
                return False
        else:
            move_content(extracted_content_path, content_path)
            return True


def move_content(old_path, new_path):
    """
        Helper function that will make sure that all
        parent folders for new path exists.

        Parameters:
            old_path (string): location of data that will be moved
            new_path (string): where the data will be saved
    """
    parent_dir = os.path.dirname(new_path)
    if not os.path.isdir(parent_dir):
        os.makedirs(parent_dir)
    if os.path.isdir(old_path):
        shutil.copytree(old_path, new_path, dirs_exist_ok=True)
    else:
        shutil.move(old_path, new_path)



# vim: syntax=python tabstop=4 expandtab
# coding: utf-8

import re

from hydra_genetics.utils.models.hotspot import MultiBpVariant


class Reader(object):
    def __init__(self, filename, compressed=None, encoding='ascii'):
        super().__init__()

        if not filename:
            raise Exception("A filepath needs to be specified: %" % filepath)

        if compressed:
            self._reader = codecs.getreader(encoding)(self._reader)
        else:
            if compressed is None:
                compressed = filename.endswith('.gz')
        self._reader = open(filename, 'rb' if compressed else 'rt')
        self.filename = filename

        self.reader = (line.rstrip() for line in self._reader)

        self._row_pattern = re.compile("\t")

    def __del__(self):
        if self._reader and not self._reader.closed:
            self._reader.close()

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        '''Return the next record in the file.'''
        line = next(self.reader)
        row = self._row_pattern.split(line.rstrip())
        chrom = row[0]
        start = row[1]
        stop = row[2]
        ref = row[3]
        var = row[4]
        gene = row[5]
        cds_change = row[6]
        aa_change = row[7]
        transcript = row[8]

        return MultiBpVariant(chrom, start, stop, ref, var, gene, cds_change, aa_change, transcript)
